/*--------------------------------------------------------------*/
/*  qrouter.c -- general purpose autorouter                     */
/*  uses magic to automatically generate layout			*/
/*--------------------------------------------------------------*/
/* Written by Tim Edwards, June 2011, based on code by Steve	*/
/* Beccue, 2003							*/
/*--------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "qrouter.h"
#include "config.h"
#include "node.h"
#include "maze.h"
#include "lef.h"

int  Pathon = -1;
int  TotalRoutes = 0;

FILE *Failfptr = NULL;
FILE *CNfptr = NULL;

NET     Nlnets;		// list of nets in the design
NET	CurNet;		// current net to route, used by 2nd stage
STRING  DontRoute;      // a list of nets not to route (e.g., power)
STRING  CriticalNet;    // list of critical nets to route first
GATE    GateInfo;       // standard cell macro information
GATE    Nlgates;	// gate instance information
NODE    Nlnodes;	// information about gate terminals (pins)
NETLIST FailedNets;	// list of nets that failed to route
NETLIST Abandoned;	// list of nets that will never route

u_char *Mask[MAX_LAYERS];    // mask out best area to route, expand as needed
u_int  *Obs[MAX_LAYERS];     // net obstructions in layer
u_int  *Obs2[MAX_LAYERS];    // used for pt->pt routes on layer
float  *Stub[MAX_LAYERS];    // used for stub routing to pins
NODE   *Nodeloc[MAX_LAYERS]; // nodes are here. . .
NODE   *Nodesav[MAX_LAYERS]; // . . . and here (but not to be altered)
PROUTE Pr[PRindMAX + 1];     // put this in the Obs2 array
DSEG   UserObs;		     // user-defined obstruction layers

int   Numnodes = 0;
int   Numnets = 0;
int   Numgates = 0;
int   Numpins = 0;
int   Verbose = 0;
int   PRind = 0;
int   Debug = 0;                 // print level for debug information
int   CurrentPass = 0;           // Current Pass

/*--------------------------------------------------------------*/
/* Open the "failed" and "cn" (critical nets) files.		*/
/*--------------------------------------------------------------*/

void openFailFile()
{
   Failfptr = fopen("failed", "w");
   if (!Failfptr) {
      fprintf(stderr, "Could not open file \"failed\"\n");
   }

   CNfptr = fopen("cn", "w");
   if (!CNfptr) {
      fprintf(stderr, "Could not open file \"cn\".\n" );
   }
}
    
/*--------------------------------------------------------------*/
/* main - program entry point, parse command line		*/
/*								*/
/*   ARGS: argc (count) argv, command line 			*/
/*   RETURNS: to OS						*/
/*   SIDE EFFECTS: 						*/
/*--------------------------------------------------------------*/

main(argc, argv)
   int	argc;
   char	*argv[];
{
   int	i, j;
   int length, width;
   FILE *l, *configFILEptr, *fptr;
   u_int u;
   static char configdefault[] = CONFIGFILENAME;
   char *configfile = configdefault;
   char *infofile = NULL;
   char *dotptr, *sptr;
   char DEFfilename[256];
   char Filename[256];

   NET net;
    
   Filename[0] = 0;
   DEFfilename[0] = 0;

   while ((i = getopt(argc, argv, "c:d:i:hv:")) != -1) {
      switch (i) {
	 case 'c':
	    configfile = strdup(optarg);
	    break;
	 case 'd':
	    Debug = atoi(optarg);
	    break;
	 case 'v':
	    Verbose = atoi(optarg);
	    break;
	 case 'i':
	    infofile = strdup(optarg);
	    break;
	 default:
	    fprintf(stderr, "bad switch %d\n", i);
	 case 'h':
	    helpmessage();
	    exit(0);
	    break;
      }
   }

   configFILEptr = fopen(configfile, "r");

   if (!configFILEptr) {
      char *configtmp;
      configtmp = malloc(strlen(configfile) + strlen(LIBDIR) + 2);
      sprintf(configtmp, "%s/%s", LIBDIR, configfile);
      if (configfile != configdefault) free(configfile);
      configfile = configtmp;
      configFILEptr = fopen(configfile, "r" );
   }

   if (configFILEptr) {
       read_config(configFILEptr);
   }
   else {
      fprintf(stderr, "Could not open %s!\n", configfile );
   }
   if (configfile != configdefault) free(configfile);

   if (infofile != NULL) {
      FILE *infoFILEptr;

      infoFILEptr = fopen(infofile, "w" );
      if (infoFILEptr != NULL) {
	 // fprintf(infoFILEptr, "INFO qrouter %s\n", QVERSION);
      
         /* Print information about route layers, and exit */
         for (i = 0; i < Num_layers; i++) {
	    fprintf(infoFILEptr, "%s %g %g %g %s\n",
			LefGetRouteName(i), LefGetRoutePitch(i),
			LefGetRouteOffset(i), LefGetRouteWidth(i),
			(LefGetRouteOrientation(i) == 1) ? "horizontal"
			: "vertical");
	 }
	 fclose(infoFILEptr);
      }
      exit(0);
   }

   if (optind < argc) {

      /* process remaining commandline strings */

      strcpy( Filename, argv[optind] );
      dotptr = strrchr(Filename, '.');
      if (dotptr != NULL) *dotptr = '\0';
      sprintf(DEFfilename, "%s.def", Filename);
   }
   else {
      fprintf(stderr, "No netlist file!\n");
      helpmessage();
      exit(1);
   }

   DefRead(DEFfilename);
   create_netorder();

   for (i = 0; i < Num_layers; i++) {
      if (PitchX[i] == 0.0 || PitchY[i] == 0.0) {
	 fprintf(stderr, "Have a 0 pitch for layer %d (of %d).  "
			"Exit.\n", i + 1, Num_layers);
	 exit(-3);
      }
      NumChannelsX[i] = (int)(1.5 + (Xupperbound - Xlowerbound) / PitchX[i]);
      NumChannelsY[i] = (int)(1.5 + (Yupperbound - Ylowerbound) / PitchY[i]);
      printf("Number of x channels for layer %d is %d\n",
		i, NumChannelsX[i]);
      printf("Number of y channels for layer %d is %d\n",
		i, NumChannelsY[i]);
	
      if (NumChannelsX[i] <= 0) {
	 fprintf(stderr, "Something wrong with layer %d x bounds.\n", i);
	 exit(-3);
      }
      if (NumChannelsY[i] <= 0) {
	 fprintf(stderr, "Something wrong with layer %d y bounds.\n", i);
	 exit(-3);
      }
      fflush(stdout);

      /*
      Mask[i] = (u_char *)calloc(NumChannelsX[i] * NumChannelsY[i],
			sizeof(u_char));
      if (!Mask[i]) {
	 fprintf(stderr, "Out of memory 1.\n");
	 exit(3);
      }
      */

      Obs[i] = (u_int *)calloc(NumChannelsX[i] * NumChannelsY[i],
			sizeof(u_int));
      if (!Obs[i]) {
	 fprintf(stderr, "Out of memory 2.\n");
	 exit(4);
      }

      Obs2[i] = (u_int *)calloc(NumChannelsX[i] * NumChannelsY[i],
			sizeof(u_int));
      if (!Obs2[i]) {
	 fprintf( stderr, "Out of memory 3.\n");
	 exit(5);
      }

      Stub[i] = (float *)calloc(NumChannelsX[i] * NumChannelsY[i],
			sizeof(float));
      if (!Stub[i]) {
	 fprintf( stderr, "Out of memory 4.\n");
	 exit(6);
      }

      // Nodeloc is the reverse lookup table for nodes

      Nodeloc[i] = (NODE *)calloc(NumChannelsX[i] * NumChannelsY[i],
		sizeof(NODE));
      if (!Nodeloc[i]) {
         fprintf(stderr, "Out of memory 5.\n");
         exit(7);
      }

      Nodesav[i] = (NODE *)calloc(NumChannelsX[i] * NumChannelsY[i],
		sizeof(NODE));
      if (!Nodesav[i]) {
         fprintf(stderr, "Out of memory 6.\n");
         exit(8);
      }
   }
   fflush(stdout);

   fprintf(stderr, "Diagnostic: memory block is %d bytes\n",
		sizeof(u_int) * NumChannelsX[0] * NumChannelsY[0]);

   /* Be sure to create obstructions from gates first, since we don't	*/
   /* want improperly defined or positioned obstruction layers to over-	*/
   /* write our node list.						*/

   create_obstructions_from_gates();
   create_obstructions_from_nodes();
   create_obstructions_from_variable_pitch();
    
   // now we have netlist data, and can use it to get a list of nets.

   FailedNets = (NETLIST)NULL;
   Abandoned = (NETLIST)NULL;
   fflush(stdout);
   fprintf(stderr, "Numnets = %d, Numnodes = %d, Numpins = %d\n",
	     Numnets - MIN_NET_NUMBER, Numnodes, Numpins );

   // print_nlgates( "net.details" );
   // print_nodes( "nodes.details" );
   // print_nlnets( "netlist.out" );
   // print_obs( "obs.out" );

   for (i = 0; i < Numnets; i++) {
      net = getnettoroute(i);
      if (net != NULL) doroute(net, (u_char)0);
   }

   fflush(stdout);
   fprintf(stdout, "\n----------------------------------------------\n");
   fprintf(stdout, "Progress: ");
   fprintf(stdout, "Total routing loops completed: %d\n", TotalRoutes);
   if (FailedNets == (NETLIST)NULL && Abandoned == (NETLIST)NULL)
      fprintf(stdout, "No failed routes!\n");
   else {
      if (FailedNets != (NETLIST)NULL)
          fprintf(stdout, "Failed net routes: %d\n", FailedNets->idx);
      if (Abandoned != (NETLIST)NULL)
          fprintf(stdout, "Abandoned net routes: %d\n", Abandoned->idx);
   }
   fprintf(stdout, "----------------------------------------------\n");

   dosecondstage();

   // Finish up by writing the routes to an annotated DEF file
    
   emit_routes(DEFfilename);

   fprintf(stdout, "----------------------------------------------\n");
   fprintf(stdout, "Final: ");
   if (FailedNets == (NETLIST)NULL && Abandoned == (NETLIST)NULL)
      fprintf(stdout, "No failed routes!\n");
   else {
      if (FailedNets != (NETLIST)NULL) {
         fprintf(stdout, "Failed net routes: %d\n", FailedNets->idx);
         fprintf(stdout, "See file \"failed\" for list of failing nets.\n");
      }
      if (Abandoned != (NETLIST)NULL) {
         fprintf(stdout, "Abandoned net routes: %d\n", Abandoned->idx);
         fprintf(stdout, "See file \"failed\" for list of abandoned nets.\n");
      }
   }
   fprintf(stdout, "----------------------------------------------\n");

   exit(0);

} /* main() */

/*--------------------------------------------------------------*/
/* pathstart - begin a DEF format route path           		*/
/*								*/
/* 	If "special" is true, then this path is in a		*/
/*	SPECIALNETS section, in which each route specifies	*/
/*	a width.						*/
/*--------------------------------------------------------------*/

void pathstart(FILE *cmd, int layer, int x, int y, u_char special)
{
   if (Pathon == 1) {
      fprintf( stderr, "pathstart():  Major error.  Started a new "
		"path while one is in progress!\n"
		"Doing it anyway.\n" );
   }

   if (layer >= 0) {
      if (Pathon == -1)
	 fprintf(cmd, "+ ROUTED ");
      else
	 fprintf(cmd, "\n  NEW ");
      if (special)
         fprintf(cmd, "%s %d ( %d %d ) ", CIFLayer[layer],
			(int)(100 * LefGetViaWidth(layer, layer, 0) + 0.5),
			x, y);
      else
         fprintf(cmd, "%s ( %d %d ) ", CIFLayer[layer], x, y);
   }
   Pathon = 1;

} /* pathstart() */

/*--------------------------------------------------------------*/
/* pathto  - continue a path to the next point        		*/
/*								*/
/*   ARGS: coordinate pair					*/
/*   RETURNS: 							*/
/*   SIDE EFFECTS: 						*/
/*--------------------------------------------------------------*/

void pathto(FILE *cmd, int x, int y, int horizontal, int vertical)
{
    if (Pathon <= 0) {
	fprintf(stderr, "pathto():  Major error.  Added to a "
		"non-existent path!\n"
		"Doing it anyway.\n");
    }
    fprintf(cmd, "( ");
    if (vertical)
	fprintf(cmd, "* ");
    else
	fprintf(cmd, "%d ", x);

    if (horizontal)
	fprintf(cmd, "* ");
    else
	fprintf(cmd, "%d ", y);

    fprintf(cmd, ") ");

} /* pathto() */

/*--------------------------------------------------------------*/
/* pathvia  - add a via to a path               		*/
/*								*/
/*   ARGS: coord						*/
/*   RETURNS: 							*/
/*   SIDE EFFECTS: 						*/
/*--------------------------------------------------------------*/

void pathvia(FILE *cmd, int layer, int x, int y)
{
   char *s;

  s = Via[layer];
  if (Pathon <= 0) {
     if (Pathon == -1)
	fprintf(cmd, "+ ROUTED ");
     else 
	fprintf(cmd, "\n  NEW ");
     fprintf(cmd, "%s ( %d %d ) ", CIFLayer[layer], x, y);
  }
  fprintf(cmd, "%s ", s);
  Pathon = 0;

} /* pathvia() */

/*--------------------------------------------------------------*/
/* print_nets - print the nets list - created from Nlgates list */
/*								*/
/*   ARGS: filename to list to					*/
/*   RETURNS: nothing						*/
/*   SIDE EFFECTS: 						*/
/*   AUTHOR and DATE: steve beccue      Sat July 26		*/
/*--------------------------------------------------------------*/

void print_nets(char *filename)
{
   FILE *o;
   GATE g;
   int i;
   DSEG drect;

   if (!strcmp(filename, "stdout")) {
	o = stdout;
   } else {
	o = fopen(filename, "w");
   }
   if (!o) {
	fprintf(stderr, "route:print_nets.  Couldn't open output file\n");
	return;
   }

   for (g = Nlgates; g; g = g->next) {
      fprintf(o, "%s: %s: nodes->", g->gatename, g->gatetype);
      for (i = 0; i < g->nodes; i++) {
	 // This prints the first tap position only.
	 drect = g->taps[i];
	 fprintf( o, "%s(%g,%g) ", g->node[i], drect->x1, drect->y1);
      }
   }
   fprintf( o, "\n");
} /* print_nets() */

/*--------------------------------------------------------------*/
/* print_routes - print the routes list				*/
/*								*/
/*   ARGS: filename to list to					*/
/*   RETURNS: nothing						*/
/*   SIDE EFFECTS: 						*/
/*   AUTHOR and DATE: steve beccue      Sat July 26		*/
/*--------------------------------------------------------------*/

void print_routes( char *filename )
{
    FILE *o;
    GATE g;
    int i;

    if( !strcmp( filename, "stdout" ) ) {
	o = stdout;
    } else {
	o = fopen( filename, "w" );
    }
    if( !o ) {
	fprintf( stderr, "route:print_routes.  Couldn't open output file\n" );
	return;
    }

    for (g = Nlgates; g; g = g->next) {
	fprintf( o, "%s: %s: nodes->", g->gatename, g->gatetype );
	for( i = 0 ; i < g->nodes; i++ ) {
	    fprintf( o, "%s ", g->node[i] );
	}
	fprintf(o, "\n");
    }
} /* print_routes() */

/*--------------------------------------------------------------*/
/* print_nlgates - print the nlgate list			*/
/*								*/
/*   ARGS: filename to list to					*/
/*   RETURNS: nothing						*/
/*   SIDE EFFECTS: 						*/
/*   AUTHOR and DATE: steve beccue      Wed July 23		*/
/*--------------------------------------------------------------*/

void print_nlgates( char *filename )
{
    FILE *o;
    GATE g;
    int i;
    DSEG drect;

    if( !strcmp( filename, "stdout" ) ) {
	o = stdout;
    } else {
	o = fopen( filename, "w" );
    }
    if( !o ) {
	fprintf( stderr, "route:print_nlgates.  Couldn't open output file\n" );
	return;
    }

    for (g = Nlgates; g; g = g->next) {
	fprintf( o, "%s: %s: nodes->", g->gatename, g->gatetype );
	for( i = 0 ; i < g->nodes; i++ ) {
	    // This prints the first tap position only.
	    drect = g->taps[i];
	    fprintf( o, "%s(%g,%g)", g->node[i], drect->x1, drect->y1);
	}
        fprintf(o, "\n");
    }
} /* print_nlgates() */

/*--------------------------------------------------------------*/
/* getnettoroute - get a net to route				*/
/*								*/
/*   ARGS: 							*/
/*   RETURNS: 							*/
/*   SIDE EFFECTS: 						*/
/*   AUTHOR and DATE: steve beccue      Fri Aug 8		*/
/*--------------------------------------------------------------*/

NET getnettoroute(int order)
{
  NET net;
  
  fflush(stdout);
  fflush(stderr);

  for (net = Nlnets; net; net = net->next) {
     if (net->netorder == order) {
	if (net->numnodes >= 2) {
	   return net;
        }
     }
  }
  if (Verbose > 0) {
     fprintf(stderr, "getnettoroute():  Fell through\n");
  }
  return NULL;

} /* getnettoroute() */

/*--------------------------------------------------------------*/
/* dosecondstage() ---						*/
/*								*/
/* Second stage:  Rip-up and reroute failing nets.		*/
/* Method:							*/
/* 1) Route a failing net with stage = 1 (other nets become	*/
/*    costs, not blockages, no copying to Obs)			*/
/* 2) If net continues to fail, flag it as unroutable and	*/
/*    remove it from the list.					*/
/* 3) Otherwise, determine the nets with which it collided.	*/
/* 4) Remove all of the colliding nets, and add them to the	*/
/*    FailedNets list						*/
/* 5) Route the original failing net.				*/
/* 6) Continue until all failed nets have been processed.	*/
/*--------------------------------------------------------------*/

void
dosecondstage()
{
   int i, result;
   NET net;
   NETLIST nl, nl2, fn;

   while (FailedNets != NULL) {

      // Diagnostic:  how are we doing?
      i = 0;
      for (nl = FailedNets; nl; nl = nl->next) i++;
      fprintf(stdout, "------------------------------\n");
      fprintf(stdout, "Number of remaining nets: %d\n", i);
      fprintf(stdout, "------------------------------\n");

      net = FailedNets->net;

      // Remove this net from the fail list
      nl2 = FailedNets;
      FailedNets = FailedNets->next;
      free(nl2);

      // Route as much as possible without collisions
      result = doroute(net, (u_char)0);

      if (result != 0) {
	 fprintf(stderr, "Routing net with collisions\n");
         result = doroute(net, (u_char)1);
         if (result != 0) {
	    if (net->noripup != NULL) {
	       if ((net->flags & NET_PENDING) == 0) {
	          // Clear this net's "noripup" list and try again.

	          while (net->noripup) {
	             nl = net->noripup->next;
	             free(net->noripup);
	             net->noripup = nl;
	          }
	          result = doroute(net, (u_char)1);
		  net->flags |= NET_PENDING;	// Next time we abandon it.
	       }
	    }
	 }
         if (result != 0) {
	    // Complete failure to route, even allowing collisions.
	    // Abandon routing this net.
	    fprintf(stderr, "----------------------------------------------\n");
	    fprintf(stderr, "Complete failure on net %s:  Abandoning.\n",
			net->netname);
	    fprintf(stderr, "----------------------------------------------\n");
	    // Add the net to the "abandoned" list
	    nl = (NETLIST)malloc(sizeof(struct netlist_));
	    nl->net = net;
	    if (Abandoned != (NETLIST)NULL)
	       nl->idx = Abandoned->idx + 1;
	    else
	       nl->idx = 1;
	    nl->next = Abandoned;
	    Abandoned = nl;

	    while (FailedNets && (FailedNets->net == net)) {
	       nl = FailedNets->next;
	       free(FailedNets);
	       FailedNets = nl;
	    }
	    continue;
	 }

         // Analyze route for nets with which it collides

         nl = find_colliding(net);
      }
      else
	 nl = (NETLIST)NULL;

      // Remove the colliding nets from the route grid and append
      // them to FailedNets.

      while(nl) {
	 nl2 = nl->next;
         fprintf(stdout, "Ripping up blocking net %s\n", nl->net->netname);
	 ripup_net(nl->net, (u_char)1);
	 for (fn = FailedNets; fn && fn->next != NULL; fn = fn->next);
	 if (fn)
	    fn->next = nl;
	 else
	    FailedNets = nl;

	 // Add nl->net to "noripup" list for this net, so it won't be
	 // routed over again by the net.  Avoids infinite looping in
	 // the second stage.

	 fn = (NETLIST)malloc(sizeof(struct net_));
	 fn->next = net->noripup;
	 net->noripup = fn;
	 fn->net = nl->net;

	 nl->next = (NETLIST)NULL;
	 nl = nl2;
      }

      // Now we copy the net we routed above into Obs
      writeback_all_routes(net);
   }

   // Make sure FailedNets is cleaned up, in case we broke out
   // of the loop with an error condition.

   if (Failfptr) fprintf(Failfptr, "\n------------------------------------------\n");
   if (FailedNets && !Failfptr) openFailFile();

   while (FailedNets) {
      net = FailedNets->net;
      fprintf(Failfptr, "Final:  Failed to route net %s\n", net->netname);
      nl = FailedNets->next;
      free(FailedNets);
      FailedNets = nl;
   }
}

/*--------------------------------------------------------------*/
/* createMask() ---						*/
/*								*/
/* Create mask of optimal area to route.			*/
/* For 2-node routes, find the two L-shaped routes between the	*/
/* two closest points of the nodes.				*/
/* For multi-node (>2) routes, find the best trunk line that	*/
/* passes close to all nodes, and generate stems to the closest	*/
/* point on each node.						*/
/*--------------------------------------------------------------*/

void createMask(NET net)
{
  NODE n1, n2;
  NODELIST n;
  int i, j, o, l;
  DPOINT dtap, d1tap, d2tap, mintap;
  int dx, dy, dist, mindist;
  int x1, x2, y1, y2;
  int xcent, ycent, xmin, ymin, xmax, ymax;

  fillMask(0);

  if (net->numnodes == 2) { 

     n1 = (NODE)net->netnodes->node;
     n2 = (NODE)net->netnodes->next->node;
     mindist = MAXRT;

     // Simple 2-pass---pick up first tap on n1, find closest tap on n2,
     // then find closest tap on n1.
     d1tap = (n1->taps == NULL) ? n1->extend : n1->taps;
     for (d2tap = (n2->taps == NULL) ? n2->extend : n2->taps; d2tap != NULL;
		d2tap = d2tap->next) {
	dx = d2tap->gridx - d1tap->gridx;
	dy = d2tap->gridy - d1tap->gridy;
	dist = dx * dx + dy * dy;
	if (dist < mindist) {
	   mindist = dist;
	   mintap = d2tap;
	}
     }
     d2tap = mintap;
     mindist = MAXRT;
     for (d1tap = (n1->taps == NULL) ? n1->extend : n1->taps; d1tap != NULL;
		d1tap = d1tap->next) {
	dx = d2tap->gridx - d1tap->gridx;
	dy = d2tap->gridy - d1tap->gridy;
	dist = dx * dx + dy * dy;
	if (dist < mindist) {
	   mindist = dist;
	   mintap = d1tap;
	}
     }
     d1tap = mintap;

     // Now construct a mask of two L-routes from d1tap to d2tap, 3 tracks width

     x1 = (d1tap->gridx < d2tap->gridx) ? d1tap->gridx : d2tap->gridx;
     x2 = (d1tap->gridx < d2tap->gridx) ? d2tap->gridx : d1tap->gridx;
     y1 = (d1tap->gridy < d2tap->gridy) ? d1tap->gridy : d2tap->gridy;
     y2 = (d1tap->gridy < d2tap->gridy) ? d2tap->gridy : d1tap->gridy;

     l = (d1tap->layer < d2tap->layer) ? d1tap->layer : d2tap->layer;

     // Place a track on every tap and extend position
     for (d1tap = n1->taps; d1tap != NULL; d1tap = d1tap->next)
	Mask[d1tap->layer][OGRID(d1tap->x, d1tap->y, d1tap->layer)] = (u_char)1;
     for (d1tap = n1->extend; d1tap != NULL; d1tap = d1tap->next)
	Mask[d1tap->layer][OGRID(d1tap->x, d1tap->y, d1tap->layer)] = (u_char)1;
     for (d2tap = n2->taps; d2tap != NULL; d2tap = d2tap->next)
	Mask[d2tap->layer][OGRID(d2tap->x, d2tap->y, d2tap->layer)] = (u_char)1;
     for (d2tap = n2->extend; d2tap != NULL; d2tap = d2tap->next)
	Mask[d2tap->layer][OGRID(d2tap->x, d2tap->y, d2tap->layer)] = (u_char)1;

     // Find the orientation of the lowest tap layer.  Lay alternate vertical
     // and horizontal tracks according to track orientation.

     for (; l < Num_layers; l++) {
        o = LefGetRouteOrientation(l);
	if (!o) {
           for (i = x1 - 1; i <= x1 + 1; i++)
	      for (j = y1 - 1; j <= y2 + 1; j++)	// Left vertical route
	         Mask[l][OGRID(i, j, l)] = (u_char)1;

           for (i = x2 - 1; i <= x2 + 1; i++)
	      for (j = y1 - 1; j <= y2 + 1; j++)	// Right vertical route
	         Mask[l][OGRID(i, j, l)] = (u_char)1;
	}
	else {
           for (i = x1 - 1; i <= x2 + 1; i++)
	      for (j = y1 - 1; j <= y1 + 1; j++)	// Bottom horizontal route
	         Mask[l][OGRID(i, j, l)] = (u_char)1;

           for (i = x1 - 1; i <= x2 + 1; i++)
	      for (j = y2 - 1; j <= y2 + 1; j++)	// Top horizontal route
	         Mask[l][OGRID(i, j, l)] = (u_char)1;
	}
     }

     fprintf(stdout, "2-port mask edges rectangle (%d %d) to (%d %d)\n",
		x1, y1, x2, y2);
  }
  else {

     // Use the first tap point for each node to get a rough bounding box and
     // centroid of all taps
     xcent = ycent = 0;
     xmax = ymax = -(MAXRT);
     xmin = ymin = MAXRT;
     for (n = net->netnodes; n != NULL; n = n->next) {
	n1 = n->node;
	dtap = (n1->taps == NULL) ? n1->extend : n1->taps;
	xcent += dtap->gridx;
	ycent += dtap->gridy;
	if (dtap->gridx > xmax) xmax = dtap->gridx;
	if (dtap->gridx < xmin) xmin = dtap->gridx;
	if (dtap->gridy > ymax) ymax = dtap->gridy;
	if (dtap->gridy < ymin) ymin = dtap->gridy;
     }
     xcent /= net->numnodes;
     ycent /= net->numnodes;

     if (xmax - xmin > ymax - ymin) {
	// Horizontal trunk
	o = 1;
	ymin = ymax = ycent;
     }
     else {
	o = 0;
	xmin = xmax = xcent;
     }

     // Allow routes at all tap and extension points
     for (n = net->netnodes; n != NULL; n = n->next) {
	n1 = n->node;
        for (dtap = n1->taps; dtap != NULL; dtap = dtap->next)
	   Mask[dtap->layer][OGRID(dtap->gridx, dtap->gridy, dtap->layer)] = (u_char)1;
        for (dtap = n1->extend; dtap != NULL; dtap = dtap->next)
	   Mask[dtap->layer][OGRID(dtap->gridx, dtap->gridy, dtap->layer)] = (u_char)1;
     }

     for (l = 0; l < Num_layers; l++) {
	// If the layer orientation is the same as the trunk, place mask around
	// trunk line.
	if (o == LefGetRouteOrientation(l)) {
           for (i = xmin - 1; i <= xmax + 1; i++)
	      for (j = ymin - 1; j <= ymax + 1; j++)
	         Mask[l][OGRID(i, j, l)] = (u_char)1;
	}
	else {
           for (n = net->netnodes; n != NULL; n = n->next) {
	      n1 = n->node;
	      dtap = (n1->taps == NULL) ? n1->extend : n1->taps;
	      if (o == 1) {	// Horizontal trunk, vertical branches
                 for (i = dtap->gridx - 1; i <= dtap->gridx + 1; i++) {
	            for (j = dtap->gridy - 1; j <= ycent + 1; j++)
	               Mask[l][OGRID(i, j, l)] = (u_char)1;
	            for (j = ycent - 1; j <= dtap->gridy + 1; j++)
	               Mask[l][OGRID(i, j, l)] = (u_char)1;
	         }
	      } 
	      else {		// Vertical trunk, horizontal branches
                 for (j = dtap->gridy - 1; j <= dtap->gridy + 1; j++) {
	            for (i = dtap->gridx - 1; i <= xcent + 1; i++)
	               Mask[l][OGRID(i, j, l)] = (u_char)1;
	            for (i = xcent - 1; i <= dtap->gridx + 1; i++)
	               Mask[l][OGRID(i, j, l)] = (u_char)1;
		 }
	      }
	   }
	}
     }
     fprintf(stdout, "multi-port mask has trunk line (%d %d) to (%d %d)\n",
		xmin, ymin, xmax, ymax);
  }
}

/*--------------------------------------------------------------*/
/* expandMask --- Extend the Mask area by 1 in all directions	*/
/* by recursive search over all active mask positions (set to	*/
/* nonzero).  Value 1 is used for new mask positions; value 2	*/
/* tracks mask positions that have been processed.		*/
/*--------------------------------------------------------------*/

void expandMask(int x, int y, int l) {
   if (Mask[l][OGRID(x, y, l)] == 0) {
      Mask[l][OGRID(x, y, l)] = (u_char)2;
   }
   else if (Mask[l][OGRID(x, y, l)] == 1) {
      Mask[l][OGRID(x, y, l)] = (u_char)2;	// Mark processed
      if (x > 0) expandMask(x - 1, y, l);
      else if (x < NumChannelsX[l] - 1) expandMask(x + 1, y, l);
      if (y > 0) expandMask(x, y - 1, l);
      else if (y < NumChannelsY[l] - 1) expandMask(x, y + 1, l);
      if (l > 0) expandMask(x, y, l - 1);
      else if (l < Num_layers - 1) expandMask(x, y, l + 1);
   }
}

/*--------------------------------------------------------------*/
/* setMask() is the same routine as expandMask() but cleans up,	*/
/* turning all values 2 back into mask value 1.			*/
/*--------------------------------------------------------------*/

void setMask(int x, int y, int l) {
   if (Mask[l][OGRID(x, y, l)] == 2) {
      Mask[l][OGRID(x, y, l)] = (u_char)1;

      if (x > 0) setMask(x - 1, y, l);
      else if (x < NumChannelsX[l] - 1) setMask(x + 1, y, l);
      if (y > 0) setMask(x, y - 1, l);
      else if (y < NumChannelsY[l] - 1) setMask(x, y + 1, l);
      if (l > 0) setMask(x, y, l - 1);
      else if (l < Num_layers - 1) setMask(x, y, l + 1);
   }
}

/*--------------------------------------------------------------*/
/* fillMask() fills the Mask[] array with all 1s as a last	*/
/* resort, ensuring that no valid routes are missed due to a	*/
/* bad guess about the optimal route positions.			*/
/*--------------------------------------------------------------*/

void fillMask(int value) {
   int i;

   for (i = 0; i < Num_layers; i++) {
      memset((void *)Mask[i], value,
		(size_t)(NumChannelsX[i] * NumChannelsY[i] * sizeof(u_char)));
   }
}

/*--------------------------------------------------------------*/
/* doroute - basic route call					*/
/*								*/
/*	stage = 0 is normal routing				*/
/*	stage = 1 is the rip-up and reroute stage		*/
/*								*/
/*   ARGS: two nodes to be connected				*/
/*   RETURNS: 0 on success, -1 on failure			*/
/*   SIDE EFFECTS: 						*/
/*   AUTHOR and DATE: steve beccue      Fri Aug 8		*/
/*--------------------------------------------------------------*/

int doroute(NET net, u_char stage)
{
  ROUTE rt1, lrt;
  NODE n1, n2;
  NETLIST nlist;
  int result;

  if (!net) {
     fprintf(stderr, "doroute():  no net to route.\n");
     return 0;
  }

  CurNet = net;				// Global, used by 2nd stage

  n1 = (NODE)(net->netnodes->node);	// Start with the first node
  while (1) {	// Keep going until we are unable to route to a terminal

     TotalRoutes++;

     rt1 = createemptyroute();
     rt1->netnum = net->netnum;
     rt1->node = n1;

     if (Verbose > 0) {
        fprintf(stdout,"doroute(): added net %d path start %d\n", 
	       n1->netnum, n1->nodenum);
     }

     // TO-DO:  When failing to route a node, we need to check if (1) all nodes
     // are routed, (2) last node is not routed, or (3) more than one node is
     // not routed.  If (3), then set n1 to the first unrouted node.  Otherwise,
     // a large and critical net may not get routed at all because one of the
     // first legs wouldn't route.

     result = route_segs(rt1, stage);

     if ((result == 0) || (rt1->node == NULL)) {
        // Nodes already routed, nothing to do
	free(rt1);
	return 0;
     }
     n1 = rt1->node;

     if (n1->routes) {
        for (lrt = n1->routes; lrt->next; lrt = lrt->next);
	lrt->next = rt1;
     }
     else {
	n1->routes = rt1;
     }

     if (result < 0) {		// Route failure.
	nlist = (NETLIST)malloc(sizeof(struct netlist_));
	nlist->net = net;
	if (FailedNets != NULL)
	   nlist->idx = FailedNets->idx + 1;
	else
	   nlist->idx = 1;
	nlist->next = FailedNets;
	FailedNets = nlist;
	return -1;
     }
  }
  
} /* doroute() */

/*--------------------------------------------------------------*/
/* route_segs - detailed route from node to node using onestep	*/
/*	method   						*/
/*								*/
/*   ARGS: ROUTE, ready to add segments to do route		*/
/*   RETURNS: NULL if failed, manhattan distance if success	*/
/*   SIDE EFFECTS: 						*/
/*   AUTHOR and DATE: steve beccue      Fri Aug 8		*/
/*--------------------------------------------------------------*/

int route_segs(ROUTE rt, u_char stage)
{
  POINT gpoint, glist, gunproc;
  NET  net;
  SEG  seg;
  struct seg_ bbox;
  int  i, j, k, o;
  int  x, y;
  NODE n1, n2, n2save;
  u_int netnum, dir;
  char filename[32];
  int  dist, max, min, maxcost;
  int  thisnetnum, thisindex, index, pass;
  int  bestX, bestY, bestL;
  int  prind;
  int  result;
  int  rval;
  u_char first = (u_char)1;
  u_char check_order[6];
  DPOINT n1tap, n2tap;

  // Make Obs2[][] a copy of Obs[][].  Convert pin obstructions to
  // terminal positions for the net being routed.

  for (i = 0; i < Num_layers; i++) {
      for (x = 0; x < NumChannelsX[i]; x++) {
	  for (y = 0; y < NumChannelsY[i]; y++) {
	      netnum = Obs[i][OGRID(x, y, i)];
	      Obs2[i][OGRID(x, y, i)] = (netnum & ~PINOBSTRUCTMASK);
	      dir = netnum & PINOBSTRUCTMASK;
	      if ((dir != 0) && (dir == STUBROUTE_X)) {
		 if ((netnum & ~PINOBSTRUCTMASK) == rt->netnum)
		    Obs2[i][OGRID(x, y, i)] = 0;	// STUBROUTE_X not routable
	      }
	  }
      }
  }

  PRind = 0;  	// Start assigning indexes after this.  Don't confuse Pr[1]
		// with source, Pr[2] with dest:  Source never gets assigned
		// a Pr[], and dest gets its own number unrelated to TARGNETNUM.

  // Pr[0] will be used to hold information about the route cost at target
  // Note that it will be placed at whichever tap has the lowest cost route.

  Pr[0].pred = 0;
  Pr[0].flags = (u_char)0;
  Pr[0].cost = MAXRT;

  // Stack of points to search
  glist = (POINT)NULL;
  gunproc = (POINT)NULL;

  // We start at the node referenced by the route structure, and set all
  // of its taps to SRCNETNUM, as well as all connected routes.

  n1 = rt->node;
  bbox.x2 = bbox.y2 = 0;
  bbox.x1 = NumChannelsX[0];
  bbox.y1 = NumChannelsY[0];
  set_node_to_net(n1, SRCNETNUM, &glist, &bbox);

  // Now search for all other nodes on the same net that have not yet been
  // routed, and set all of their taps to TARGNETNUM

  result = 0;
  for (n2 = Nlnodes; n2; n2 = n2->next) {
     if (n2 == n1) continue;
     if (n2->netnum != n1->netnum) continue;
     if (set_node_to_net(n2, TARGNETNUM, NULL, &bbox) == 0) {
	n2save = n2;
	result = 1;
     }
  }

  // Check for the possibility that there is already a route to the target
  if (!result) {
     fprintf(stdout, "Finished routing net %s\n", n1->netname);

     // Remove nodes of the net from Nodeloc so that they will not be
     // used for crossover costing of future routes.

     for (i = 0; i < Num_layers; i++) {
        for (x = 0; x < NumChannelsX[i]; x++) {
	   for (y = 0; y < NumChannelsY[i]; y++) {
	      n1 = Nodeloc[i][OGRID(x, y, i)];
	      if (n1 != (NODE)NULL)
		 if (n1->netnum == rt->netnum)
		    Nodeloc[i][OGRID(x, y, i)] = (NODE)NULL;
	   }
        }
     }

     while (glist) {
	gpoint = glist;
	glist = glist->next;
	free(gpoint);
     }
     return 0;
  }

  // Generate a search area mask representing the "likely best route".
  // createMask(CurNet);

  // Heuristic:  Set the initial cost beyond which we stop searching.
  // This value is twice the cost of a direct route across the
  // maximum extent of the source to target, divided by the square
  // root of the number of nodes in the net.  We purposely set this
  // value low.  It has a severe impact on the total run time of the
  // algorithm.  If the initial max cost is so low that no route can
  // be found, it will be doubled on each pass.

  maxcost = 2 * MAX((bbox.x2 - bbox.x1), (bbox.y2 - bbox.y1)) * SegCost +
		(int)stage * ConflictCost;
  maxcost /= (n1->numnodes - 1);

  netnum = rt->netnum;
  n2 = n2save;

  n1tap = n1->taps;
  n2tap = n2->taps;

  if ((n1tap == NULL && n1->extend == NULL) || (n2tap == NULL && n2->extend == NULL)) {
     printf("Node of net %s has no tap points---unable to route!\n", n1->netname);
     return -1;
  }
  if (n1tap == NULL) n1tap = n1->extend;
  if (n2tap == NULL) n2tap = n2->extend;

  printf("Source node @ %gum %gum layer=%d grid=(%d %d)\n",
	  n1tap->x, n1tap->y, n1tap->layer,
	  n1tap->gridx, n1tap->gridy);
  printf("Dest node @ %gum %gum layer=%d grid=(%d %d)\n",
	  n2tap->x, n2tap->y, n2tap->layer,
	  n2tap->gridx, n2tap->gridy);
  printf("netname = %s, route number %d\n", n1->netname, TotalRoutes );
  fflush(stdout);

  for (pass = 0; pass < Numpasses; pass++) {
    CurrentPass = pass;

    if (!first) {
       fprintf(stdout, "\n");
       first = (u_char)1;
    }
    fprintf(stdout, "Pass %d", pass + 1);
    fprintf(stdout, " (maxcost is %d)\n", maxcost);

    while (gpoint = glist) {

      NPX = gpoint->x1;
      NPY = gpoint->y1;
      CurrentLay = gpoint->layer;
      glist = gpoint->next;

      thisnetnum = Obs2[CurrentLay][OGRID(NPX, NPY, CurrentLay)];
      if (thisnetnum & SRCFLAG) {
	 thisindex = (thisnetnum & ~SRCFLAG);
	 thisnetnum = SRCNETNUM;
      }
      else if (thisnetnum & RTFLAG) {
         thisindex = (thisnetnum & ~RTFLAG);
      }
      else
	 thisindex = 0;

      // ignore grid positions that have already been processed
      if (thisindex && (Pr[thisindex].flags & PR_PROCESSED)) {
	 free(gpoint);
	 continue;
      }

      // if the grid position is the destination, save the position and
      // cost if minimum.

      if (thisnetnum == TARGNETNUM) {

	 if (first) {
	    fprintf(stdout, "Found a route of cost ");
	    first = (u_char)0;
	 }
	 else
	    fprintf(stdout, "|");
	 fprintf(stdout, "%d", Pr[0].cost);
	 fflush(stdout);

	 // This position may be on a route, not at a terminal, so
	 // record it.
	 bestX = NPX;
	 bestY = NPY;
	 bestL = CurrentLay;

	 // If a complete route has been found, then there's no point
	 // in searching paths with a greater cost than this one.
	 if (Pr[0].cost < maxcost) maxcost = Pr[0].cost;

         // Don't continue processing from the target
	 free(gpoint);
	 continue;
      }

      if (thisnetnum != SRCNETNUM) {

         if (Pr[thisindex].cost < MAXRT) {

	    // Severely limit the search space by not processing anything that
	    // is not under the current route mask, which identifies a narrow
	    // "best route" solution.

	    // if (Mask[CurrentLay][OGRID(NPX, NPY, CurrentLay)] == (u_char)0) {
	    //    gpoint->next = gunproc;
	    //    gunproc = gpoint;
	    //    continue;
	    // }

            // Quick check:  Limit maximum cost to limit search space
            // Move the point onto the "unprocessed" stack and we'll pick up
            // from this point on the next pass, if needed.

	    // else
            if (Pr[thisindex].cost > maxcost) {
	       gpoint->next = gunproc;
	       gunproc = gpoint;
	       continue;
	    }
	 }
      }
      free(gpoint);

      // check east/west/north/south, and bottom to top

      // 1st optimization:  Direction of route on current layer is preferred.
      o = LefGetRouteOrientation(CurrentLay);
      if (o == 1) {			// horizontal routes---check EAST and WEST first
	 check_order[0] = EAST;
	 check_order[1] = WEST;
	 check_order[2] = UP;
	 check_order[3] = DOWN;
	 check_order[4] = NORTH;
	 check_order[5] = SOUTH;
      }
      else {				// vertical routes---check NORTH and SOUTH first
	 check_order[0] = NORTH;
	 check_order[1] = SOUTH;
	 check_order[2] = UP;
	 check_order[3] = DOWN;
	 check_order[4] = EAST;
	 check_order[5] = WEST;
      }

      min = MAXRT;

      // Check order is from 0 (1st priority) to 5 (last priority).  However, this
      // is a stack system, so the last one placed on the stack is the first to be
      // pulled and processed.  Therefore we evaluate and drop positions to check
      // on the stack in reverse order (5 to 0).

      for (i = 5; i >= 0; i--) {
	 switch (check_order[i]) {
	    case EAST:
               if ((NPX + 1) < NumChannelsX[CurrentLay]) {
         	  if ((result = eval_pt(NPX + 1, NPY, CurrentLay, thisindex, stage)) == 1) {
         	     gpoint = (POINT)malloc(sizeof(struct point_));
         	     gpoint->x1 = NPX + 1;
         	     gpoint->y1 = NPY;
         	     gpoint->layer = CurrentLay;
         	     gpoint->next = glist;
         	     glist = gpoint;
                   }
               }
	       break;

	    case WEST:
               if ((NPX - 1) >= 0) {
         	  if ((result = eval_pt(NPX - 1, NPY, CurrentLay, thisindex, stage)) == 1) {
         	     gpoint = (POINT)malloc(sizeof(struct point_));
         	     gpoint->x1 = NPX - 1;
         	     gpoint->y1 = NPY;
         	     gpoint->layer = CurrentLay;
         	     gpoint->next = glist;
         	     glist = gpoint;
                  }
               }
	       break;
         
	    case SOUTH:
               if ((NPY - 1) >= 0) {
         	  if ((result = eval_pt(NPX, NPY - 1, CurrentLay, thisindex, stage)) == 1) {
         	     gpoint = (POINT)malloc(sizeof(struct point_));
         	     gpoint->x1 = NPX;
         	     gpoint->y1 = NPY - 1;
         	     gpoint->layer = CurrentLay;
         	     gpoint->next = glist;
         	     glist = gpoint;
                   }
               }
	       break;

	    case NORTH:
               if ((NPY + 1) < NumChannelsY[CurrentLay]) {
         	  if ((result = eval_pt(NPX, NPY + 1, CurrentLay, thisindex, stage)) == 1) {
         	     gpoint = (POINT)malloc(sizeof(struct point_));
         	     gpoint->x1 = NPX;
         	     gpoint->y1 = NPY + 1;
         	     gpoint->layer = CurrentLay;
         	     gpoint->next = glist;
         	     glist = gpoint;
                  }
               }
	       break;
      
	    case DOWN:
               if (CurrentLay > 0) {
         	  if ((result = eval_pt(NPX, NPY, CurrentLay - 1, thisindex, stage)) == 1) {
         	     gpoint = (POINT)malloc(sizeof(struct point_));
         	     gpoint->x1 = NPX;
         	     gpoint->y1 = NPY;
         	     gpoint->layer = CurrentLay - 1;
         	     gpoint->next = glist;
         	     glist = gpoint;
         	  }
               }
	       break;
         
	    case UP:
               if (CurrentLay < (Num_layers - 1)) {
         	  if ((result = eval_pt(NPX, NPY, CurrentLay + 1, thisindex, stage)) == 1) {
         	     gpoint = (POINT)malloc(sizeof(struct point_));
         	     gpoint->x1 = NPX;
         	     gpoint->y1 = NPY;
         	     gpoint->layer = CurrentLay + 1;
         	     gpoint->next = glist;
         	     glist = gpoint;
         	  }
               }
	       break;
            }
         }

      // Mark this node as processed
      Pr[thisindex].flags |= PR_PROCESSED;

      if (PRind >= PRindMAX) {
	 fprintf(stdout, "\n");
	 fflush(stdout);
	 fprintf(stderr, "route.c:route_segs().  PRind hit "
			"PRindMAX\nThis is a serious error\n");
	 if (!Failfptr) openFailFile();
	 fprintf(Failfptr, "# PRind hit PRindMAX\n");
	 fprintf(Failfptr, "(%g,%g) <==> (%g,%g) net=%s\tRoute=%d\n",
		     n1tap->gridx * PitchX[0], n1tap->gridy * PitchY[0], 
		     n2tap->gridx * PitchX[0], n2tap->gridy * PitchY[0],
		     n1->netname, TotalRoutes);
	 fprintf(CNfptr, "# PRind hit PRindMAX\nRoute Priority\t%s\n",
			n1->netname);
	 fflush(CNfptr);
	 fflush(Failfptr);
	    
	 while (glist) {
	    gpoint = glist;
	    glist = glist->next;
	    free(gpoint);
	 }
	 rval = -1;
	 goto done;
      }
    } // while stack is not empty

    while (glist) {
       gpoint = glist;
       glist = glist->next;
       free(gpoint);
    }

    // If we found a route, save it and return

    if (Pr[0].cost <= maxcost) {
	NPX = bestX;
	NPY = bestY;
	CurrentLay = bestL;
	Obs2[CurrentLay][OGRID(NPX, NPY, CurrentLay)] = RTFLAG;
	commit_proute(rt, stage);
	fprintf(stdout, "\nCommit to a route of cost %d\n", Pr[0].cost);
	fprintf(stdout, "Between positions (%d %d) and (%d %d)\n",
		bestX, bestY, NPX, NPY);
	rval = 1;
	goto done;	/* route success */
    }

    // Only continue loop to next pass if we failed to find a route.
    // Increase maximum cost for next pass.

    maxcost <<= 1;
    if (maxcost < 0) break;		// Overflowed unsigned integer, we're
					// probably completely hosed long before
					// this.

    if (gunproc == NULL) break;		// route failure not due to limiting
					// search to maxcost

    // Regenerate the stack of unprocessed nodes
    glist = gunproc;
    gunproc = NULL;
    
  } // pass
  
  if (!first) fprintf(stdout, "\n");
  fflush(stdout);
  fprintf(stderr, "Fell through %d passes\n", pass);
  fprintf(stderr, "(%g,%g) <==> (%g,%g) net=%s\n",
	   n1tap->x, n1tap->y, n2tap->x, n2tap->y, n1->netname);
  if (!Failfptr) openFailFile();
  fprintf(Failfptr, "(%g,%g) <==> (%g,%g) net=%s\tRoute=%d\n",
	   n1tap->x, n1tap->y, n2tap->x, n2tap->y, n1->netname, TotalRoutes);
  fprintf(CNfptr, "Route Priority\t%s\n", n1->netname);
  fflush(CNfptr);
  fflush(Failfptr);
  
  rval = -1;

done:
  
  while (gunproc) {
     gpoint = gunproc;
     gunproc = gunproc->next;
     free(gpoint);
  }
  return rval;
  
} /* route_segs() */

/*--------------------------------------------------------------*/
/* createemptyroute - begin a ROUTE structure			*/
/*								*/
/*   ARGS: a nodes						*/
/*   RETURNS: ROUTE calloc'd and ready to begin			*/
/*   SIDE EFFECTS: 						*/
/*   AUTHOR and DATE: steve beccue      Fri Aug 8		*/
/*--------------------------------------------------------------*/

ROUTE createemptyroute()
{
  ROUTE rt;

  rt = (ROUTE)calloc(1, sizeof(struct route_));
  rt->netnum = 0;
  rt->segments = (SEG)NULL;
  rt->node = (NODE)NULL;
  rt->output = FALSE;
  rt->next = (ROUTE)NULL;
  return rt;

} /* createemptyroute() */

/*--------------------------------------------------------------*/
/* emit_routed_net --						*/
/*								*/
/* Core part of emit_routes().  Dumps the DEF format for a	*/
/* complete net route to file Cmd.  If "special" is TRUE, then	*/
/* it looks only for stub routes between a grid point and an	*/
/* off-grid terminal, and dumps only the stub route geometry as	*/
/* a SPECIALNET, which takes a width parameter.  This allows	*/
/* the stub routes to be given the same width as a via, when	*/
/* the via is larger than a route width, to avoid DRC notch	*/
/* errors between the via and the terminal.  The SPECIALNETS	*/
/* are redundant;  all routing information is in the NETS	*/
/* section.  The SPECIALNETS only specify a wider route for the	*/
/* stub route.							*/
/*								*/
/* Return 1 if a stub route was encountered, so that we can	*/
/* determine if it is necessary to write a SPECIALNETS section	*/
/*--------------------------------------------------------------*/

int
emit_routed_net(FILE *Cmd, NET net, u_char special)
{
   SEG seg, saveseg, lastseg;
   ROUTE rt;
   NODE node;
   GATE g;
   u_int dir;
   int i, layer;
   int x, y, x2, y2;
   double dc;
   int horizontal, vertical;
   DPOINT dp1, dp2;
   int stubroute = 0;

   Pathon = -1;

   /* Insert routed net here */
   for (node = Nlnodes; node; node = node->next) {
      if (node->netnum == net->netnum) {
	 for (rt = node->routes; rt; rt = rt->next) {
	    if (rt->segments && !rt->output && rt->node) {
		vertical = horizontal = FALSE;

		// Check first position for terminal offsets
		seg = (SEG)rt->segments;
		lastseg = saveseg = seg;
		if (seg) {
		   dir = Obs[seg->layer][OGRID(seg->x1, seg->y1, seg->layer)];
		   dir &= PINOBSTRUCTMASK;
		   if (dir) {
		      layer = seg->layer;
		      if (special == (u_char)0)
		         fprintf(stdout, "Stub route distance %g to terminal"
				" at %d %d (%d)\n",
				Stub[layer][OGRID(seg->x1, seg->y1, layer)],
				seg->x1, seg->y1, layer);
		      stubroute = 1;
		      dc = Xlowerbound + (double)seg->x1 * PitchX[layer];
		      x = (int)((dc + 1e-4) * 100);
		      if (dir == STUBROUTE_EW)
			 dc += Stub[layer][OGRID(seg->x1, seg->y1, layer)];
		      x2 = (int)((dc + 1e-4) * 100);
		      dc = Ylowerbound + (double)seg->y1 * PitchY[layer];
		      y = (int)((dc + 1e-4) * 100);
		      if (dir == STUBROUTE_NS)
			 dc += Stub[layer][OGRID(seg->x1, seg->y1, layer)];
		      y2 = (int)((dc + 1e-4) * 100);
		      pathstart(Cmd, seg->layer, x2, y2, special);
		      if (dir == STUBROUTE_EW) {
			 vertical = FALSE;
			 horizontal = TRUE;
		      }
		      else {
			 vertical = TRUE;
			 horizontal = FALSE;
		      }
		      pathto(Cmd, x, y, horizontal, vertical);
		   }
		}

		for (seg = rt->segments; seg; seg = seg->next) {
		   layer = seg->layer;

		   // To do: pick up route layer name from lefInfo.
		   // At the moment, technology names don't even match,
		   // and are redundant between CIFLayer[] from the
		   // config file and lefInfo.

		   dc = Xlowerbound + (double)seg->x1 * PitchX[layer];
		   x = (int)((dc + 1e-4) * 100);
		   dc = Ylowerbound + (double)seg->y1 * PitchY[layer];
		   y = (int)((dc + 1e-4) * 100);
		   dc = Xlowerbound + (double)seg->x2 * PitchX[layer];
		   x2 = (int)((dc + 1e-4) * 100);
		   dc = Ylowerbound + (double)seg->y2 * PitchY[layer];
		   y2 = (int)((dc + 1e-4) * 100);
		   switch (seg->segtype) {
		      case ST_WIRE:
			 if (Pathon != 1) {	// 1st point of route seg
			    if (special == (u_char)0)
			       pathstart(Cmd, seg->layer, x, y, (u_char)0);
			    if (x == x2) {
				vertical = TRUE;
				horizontal = FALSE;
			    }
			    else if (y == y2) {
				vertical = FALSE;
				horizontal = TRUE;
			    }
			    else {
				fprintf(stderr, "Warning:  non-"
					"Manhattan wire in route\n");
			    }
			 }
			 rt->output = TRUE;
			 if (horizontal && x == x2) {
			    vertical = TRUE;
			    horizontal = FALSE;
			 }
			 if (vertical && y == y2) {
			    vertical = FALSE;
			    horizontal = TRUE;
			 }
			 if (!(x == x2) && !(y == y2)) {
			    vertical = FALSE;
			    horizontal = FALSE;
			 }
			 if (special == (u_char)0)
			    pathto(Cmd, x2, y2, horizontal, vertical);
			 break;
		      case ST_VIA:
			 rt->output = TRUE;
			 if (special == (u_char)0)
			    pathvia(Cmd, layer, x, y);
			 break;
		      default:
			 break;
		   }
		   lastseg = seg;
		}

		// Check last position for terminal offsets
		if (lastseg && ((lastseg != saveseg)
				|| (lastseg->segtype == ST_WIRE))) {
		    seg = lastseg;
		    dir = Obs[seg->layer][OGRID(seg->x2, seg->y2, seg->layer)];
		    dir &= PINOBSTRUCTMASK;
		    if (dir) {
		       layer = seg->layer;
		       if (special == (u_char)0)
		          fprintf(stdout, "Stub route distance %g to terminal"
				" at %d %d (%d)\n",
				Stub[layer][OGRID(seg->x2, seg->y2, layer)],
				seg->x2, seg->y2, layer);
		       stubroute = 1;
		       dc = Xlowerbound + (double)seg->x2 * PitchX[layer];
		       x = (int)((dc + 1e-4) * 100);
		       if (dir == STUBROUTE_EW)
			  dc += Stub[layer][OGRID(seg->x2, seg->y2, layer)];
		       x2 = (int)((dc + 1e-4) * 100);
		       dc = Ylowerbound + (double)seg->y2 * PitchY[layer];
		       y = (int)((dc + 1e-4) * 100);
		       if (dir == STUBROUTE_NS)
			  dc += Stub[layer][OGRID(seg->x2, seg->y2, layer)];
		       y2 = (int)((dc + 1e-4) * 100);
		       if (dir == STUBROUTE_EW) {
			  vertical = FALSE;
			  horizontal = TRUE;
		       }
		       else {
			  vertical = TRUE;
			  horizontal = FALSE;
		       }
		       if (Pathon != 1) pathstart(Cmd, layer, x, y, special);
		       pathto(Cmd, x2, y2, horizontal, vertical);
		    }
		}

		if (Pathon != -1) Pathon = 0;

	    } // if (rt->segments && !rt->output)
	 }
      } // if (node->routes)
   }
   return stubroute;
}

/*--------------------------------------------------------------*/
/* emit_routes - DEF file output from the Nlnodes list		*/
/*								*/
/*  Reads the <project>.def file and rewrites file		*/
/*  <project>_route.def, where each net definition has the	*/
/*  physical route appended.					*/
/*								*/
/*   ARGS: filename to list to					*/
/*   RETURNS: nothing						*/
/*   SIDE EFFECTS: 						*/
/*   AUTHOR and DATE: steve beccue      Mon Aug 11 2003		*/
/*--------------------------------------------------------------*/

void emit_routes(char *filename)
{
    FILE *Cmd;
    int i, j, numnets;
    char line[MAXLINE + 1], *lptr;
    char netname[MAX_NAME_LEN];
    NET net;
    NODE node;
    ROUTE rt;
    char newDEFfile[256];
    FILE *fdef;
    int stubroute = 0;

    fdef = fopen(filename, "r");
    if (fdef == NULL) {
       fprintf(stderr, "emit_routes(): Cannot open DEF file for reading.\n");
       return;
    } 

    if (!strcmp(filename, "stdout")) {
	Cmd = stdout;
    }
    else {
	char *dotptr;

	strcpy(newDEFfile, filename);
	dotptr = strrchr(newDEFfile, '.');
	strcpy(dotptr, "_route.def");
	Cmd = fopen(newDEFfile, "w");
    }
    if (!Cmd) {
	fprintf(stderr, "emit_routes():  Couldn't open output (routed) DEF file.\n");
	return;
    }

    // Copy DEF file up to NETS line
    numnets = 0;
    while (fgets(line, MAXLINE, fdef) != NULL) {
       lptr = line;
       while (isspace(*lptr)) lptr++;
       if (!strncmp(lptr, "NETS", 4)) {
	  sscanf(lptr + 4, "%d", &numnets);
	  break;
       }
       fputs(line, Cmd);
    }
    fputs(line, Cmd);	// Write the NETS line
    if (numnets != (Numnets - MIN_NET_NUMBER)) {
        fprintf(stderr, "emit_routes():  DEF file has %d nets, but we want"
		" to write %d\n", numnets, Numnets - MIN_NET_NUMBER);
	if (numnets > Numnets) numnets = Numnets;
    }

    for (i = 0; i < numnets; i++) {
       netname[0] == '\0';
       while (fgets(line, MAXLINE, fdef) != NULL) {
	  if ((lptr = strchr(line, ';')) != NULL) {
	     *lptr = '\n';
	     *(lptr + 1) = '\0';
	     break;
	  }
	  else {
             lptr = line;
             while (isspace(*lptr)) lptr++;
	     if (*lptr == '-') {
		lptr++;
                while (isspace(*lptr)) lptr++;
	        sscanf(lptr, "%s", netname);
	     }
	     fputs(line, Cmd);
	  }
       }

       /* Find this net */

       for (net = Nlnets; net; net = net->next) {
	  if (!strcmp(net->netname, netname))
	     break;
       }
       if (!net) {
	  fprintf(stderr, "emit_routes():  Net %s cannot be found.\n",
		netname);

	  /* Dump rest of net and continue---no routing information */
	  *(lptr) = ';';
	  fputs(line, Cmd);
	  continue;
       }
       else {
	  /* Add last net terminal, without the semicolon */
	  fputs(line, Cmd);

	  stubroute += emit_routed_net(Cmd, net, (u_char)0);
	  fprintf(Cmd, ";\n");
       }
    }

    // Finish copying the rest of the NETS section
    while (fgets(line, MAXLINE, fdef) != NULL) {
       lptr = line;
       while (isspace(*lptr)) lptr++;
       fputs(line, Cmd);
       if (!strncmp(lptr, "END", 3)) {
	  break;
       }
    }

    // If there were stub routes, repeat them in SPECIALNETS at the
    // proper width.
    if (stubroute > 0) {

       /* Reset "output" records on routes, for next pass */
       for (node = Nlnodes; node; node = node->next)
	  for (rt = node->routes; rt; rt = rt->next)
	     rt->output = 0;

       fprintf(Cmd, "\nSPECIALNETS %d ;\n", stubroute);
       fprintf(Cmd, "- stubroute1\n");
       net = Nlnets;
       for (i = 0; i < stubroute; i++) {
          for (; net; net = net->next) {
	     if (emit_routed_net(Cmd, net, (u_char)1) > 0) {
		if (i < stubroute - 1) {
		   fprintf(Cmd, " ;\n- stubroute%d\n", i + 2);
		}
		break;
	     }
	  }
       }
       fprintf(Cmd, " ;\nEND SPECIALNETS\n");
    }    

    // fprintf(Cmd, "# Autogenerated by qrouter version %s\n", VERSION);

    // Finish copying the rest of the file
    while (fgets(line, MAXLINE, fdef) != NULL) {
       fputs(line, Cmd);
    }
    fclose(fdef);
    fclose(Cmd);

} /* emit_routes() */

/*--------------------------------------------------------------*/
/* helpmessage - tell user how to use the program		*/
/*								*/
/*   ARGS: none.						*/
/*   RETURNS: nothing.						*/
/*   SIDE EFFECTS: 						*/
/*--------------------------------------------------------------*/

void helpmessage()
{
    fprintf(stdout, "qrouter - maze router by Tim Edwards\n\n");
    fprintf(stdout, "usage:  qrouter [-switchs] design_name\n\n");
    fprintf(stdout, "switches:\n");
    fprintf(stdout, "\t-c <file>\t\t\tConfiguration file name if not route.cfg.\n");
    fprintf(stdout, "\t-v <level>\t\t\tVerbose output level.\n");
    fprintf(stdout, "\t-i <file>\t\t\tPrint route names and pitches and exit.\n");
    // fprintf(stdout, "\t-d\t\t\tDebug file output.\n");
    fprintf(stdout, "\n");
    fprintf(stdout, "%s\n", VERSION);

} /* helpmessage() */

/* end of qrouter.c */
